from __future__ import annotations

import calendar
from datetime import date, datetime, time
from datetime import timedelta as td
from datetime import timezone
from enum import IntEnum
from typing import Set, Tuple, Union, cast

UTC = timezone.utc

SpecItem = Union[int, Tuple[int, int]]

RANGES = [
    range(0, 60),
    range(0, 24),
    range(1, 32),
    range(1, 13),
    range(0, 8),
]
SYMBOLIC_DAYS = "SUN MON TUE WED THU FRI SAT".split()
SYMBOLIC_MONTHS = "JAN FEB MAR APR MAY JUN JUL AUG SEP OCT NOV DEC".split()
DAYS_IN_MONTH = [-1, 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]
FIELD_NAMES = ["minute", "hour", "day-of-month", "month", "day-of-week"]


class CronSimError(Exception):
    pass


class Field(IntEnum):
    MINUTE = 0
    HOUR = 1
    DAY = 2
    MONTH = 3
    DOW = 4

    def msg(self) -> str:
        return "Bad %s" % FIELD_NAMES[self]

    def _int(self, value: str) -> int:
        if value == "":
            raise CronSimError(self.msg())
        for ch in value:
            if ch not in "0123456789":
                raise CronSimError(self.msg())

        return int(value)

    def int(self, s: str) -> int:
        if self == Field.MONTH and s in SYMBOLIC_MONTHS:
            return SYMBOLIC_MONTHS.index(s) + 1

        if self == Field.DOW and s in SYMBOLIC_DAYS:
            return SYMBOLIC_DAYS.index(s)

        v = self._int(s)
        if v not in RANGES[self]:
            raise CronSimError(self.msg())

        return v

    def parse(self, s: str) -> Set[SpecItem]:
        if s == "*":
            return set(RANGES[self])

        if "," in s:
            result = set()
            for term in s.split(","):
                result.update(self.parse(term))
            return result

        if self == Field.DOW and "L" in s:
            value = s[:-1]
            if not value.isdigit():
                raise CronSimError(self.msg())

            dow = self.int(s[:-1])
            return {(dow, CronSim.LAST)}

        if "#" in s and self == Field.DOW:
            term, nth_str = s.split("#", maxsplit=1)
            nth = self._int(nth_str)
            if nth < 1 or nth > 5:
                raise CronSimError(self.msg())

            nth_tuple = (self.int(term), nth)
            return {nth_tuple}

        if "/" in s:
            term, step_str = s.split("/", maxsplit=1)
            step = self._int(step_str)
            if step == 0:
                raise CronSimError(self.msg())

            items = self.parse(term)
            if items == {CronSim.LAST} or items == {CronSim.LAST_WEEKDAY}:
                return items

            if len(items) == 1:
                start = items.pop()
                assert isinstance(start, int)
                end = max(RANGES[self])
                tail = range(start, end + 1)
                return set(tail[::step])

            # items is an unordered set, so sort it before taking
            # every step-th item. Then convert it back to set.
            return set(sorted(items)[::step])

        if "-" in s:
            start_str, end_str = s.split("-", maxsplit=1)
            start = self.int(start_str)
            end = self.int(end_str)

            if end < start:
                raise CronSimError(self.msg())

            return set(range(start, end + 1))

        if self == Field.DAY and s == "LW":
            return {CronSim.LAST_WEEKDAY}

        if self == Field.DAY and s == "L":
            return {CronSim.LAST}

        return {self.int(s)}


def is_imaginary(dt: datetime) -> bool:
    return dt != dt.astimezone(UTC).astimezone(dt.tzinfo)


def last_weekday(year: int, month: int) -> int:
    """Return the date of the last weekday of a given year and month."""
    first_dow, last_date = calendar.monthrange(year, month)
    last_dow = (first_dow + last_date - 1) % 7

    if last_dow == 6:
        return last_date - 2
    elif last_dow == 5:
        return last_date - 1

    return last_date


class CronSim(object):
    LAST = -1000
    LAST_WEEKDAY = -1001

    def __init__(self, expr: str, dt: datetime):
        self.dt = dt.replace(second=0, microsecond=0)

        self.parts = expr.upper().split()
        if len(self.parts) != 5:
            raise CronSimError("Wrong number of fields")

        # In Debian cron, if either the day-of-month or the day-of-week field
        # starts with a star, then there is an "AND" relationship between them.
        # Otherwise it's "OR".
        self.day_and = self.parts[2].startswith("*") or self.parts[4].startswith("*")

        self.minutes = cast(Set[int], Field.MINUTE.parse(self.parts[0]))
        self.hours = cast(Set[int], Field.HOUR.parse(self.parts[1]))
        self.days = cast(Set[int], Field.DAY.parse(self.parts[2]))
        self.months = cast(Set[int], Field.MONTH.parse(self.parts[3]))
        self.weekdays = Field.DOW.parse(self.parts[4])

        if len(self.days) and min(self.days) > 29:
            # Check if we have any month with enough days
            if min(self.days) > max(DAYS_IN_MONTH[month] for month in self.months):
                raise CronSimError(Field.DAY.msg())

        self.fixup_tz = None
        if self.dt.tzinfo in (None, UTC):
            # No special DST handling for UTC
            pass
        else:
            if not self.parts[0].startswith("*") and not self.parts[1].startswith("*"):
                # Will use special handling for jobs that run at specific time, or
                # with a granularity greater than one hour (to mimic Debian cron).
                self.fixup_tz = self.dt.tzinfo
                self.dt = self.dt.replace(tzinfo=None)

    def tick(self, minutes: int = 1) -> None:
        """Roll self.dt forward by 1 or more minutes and fix timezone."""

        if self.dt.tzinfo not in (None, UTC):
            as_utc = self.dt.astimezone(UTC)
            as_utc += td(minutes=minutes)
            self.dt = as_utc.astimezone(self.dt.tzinfo)
        else:
            self.dt += td(minutes=minutes)

    def advance_minute(self) -> bool:
        """Roll forward the minute component until it satisfies the constraints.

        Return False if the minute meets contraints without modification.
        Return True if self.dt was rolled forward.

        """

        if self.dt.minute in self.minutes:
            return False

        if len(self.minutes) == 1:
            # An optimization for the special case where self.minutes has exactly
            # one element. Instead of advancing one minute per iteration,
            # make a jump from the current minute to the target minute.
            delta = (next(iter(self.minutes)) - self.dt.minute) % 60
            self.tick(minutes=delta)

        while self.dt.minute not in self.minutes:
            self.tick()
            if self.dt.minute == 0:
                # Break out to re-check month, day and hour
                break

        return True

    def advance_hour(self) -> bool:
        """Roll forward the hour component until it satisfies the constraints.

        Return False if the hour meets contraints without modification.
        Return True if self.dt was rolled forward.

        """

        if self.dt.hour in self.hours:
            return False

        self.dt = self.dt.replace(minute=0)
        while self.dt.hour not in self.hours:
            self.tick(minutes=60)
            if self.dt.hour == 0:
                # break out to re-check month and day
                break

        return True

    def match_dom(self, d: date) -> bool:
        """Return True is day-of-month matches."""
        if d.day in self.days:
            return True

        # Optimization: there are no months with fewer than 28 days.
        # If 28th is Sunday, the last weekday of the month is the 26th.
        # Any date before 26th cannot be the the last weekday of the month.
        if self.LAST_WEEKDAY in self.days and d.day >= 26:
            if d.day == last_weekday(d.year, d.month):
                return True

        # Optimization: there are no months with fewer than 28 days,
        # so any date before 28th cannot be the the last day of the month
        if self.LAST in self.days and d.day >= 28:
            _, last = calendar.monthrange(d.year, d.month)
            if d.day == last:
                return True

        return False

    def match_dow(self, d: date) -> bool:
        """Return True is day-of-week matches."""
        dow = d.weekday() + 1
        if dow in self.weekdays or dow % 7 in self.weekdays:
            return True

        if (dow, self.LAST) in self.weekdays or (dow % 7, self.LAST) in self.weekdays:
            _, last = calendar.monthrange(d.year, d.month)
            if d.day + 7 > last:
                # Same day next week would be outside this month.
                # So this is the last one this month.
                return True

        idx = (d.day + 6) // 7
        if (dow, idx) in self.weekdays or (dow % 7, idx) in self.weekdays:
            return True

        return False

    def match_day(self, d: date) -> bool:
        if self.day_and:
            return self.match_dom(d) and self.match_dow(d)

        return self.match_dom(d) or self.match_dow(d)

    def advance_day(self) -> bool:
        """Roll forward the day component until it satisfies the constraints.

        This method advances the date until it matches either the
        day-of-month, or the day-of-week constraint.

        Return False if the day meets contraints without modification.
        Return True if self.dt was rolled forward.

        """

        needle = self.dt.date()
        if self.match_day(needle):
            return False

        while not self.match_day(needle):
            needle += td(days=1)
            if needle.day == 1:
                # We're in a different month now, break out to re-check month
                # This significantly speeds up the "0 0 * 2 MON#5" case
                break

        self.dt = datetime.combine(needle, time(), tzinfo=self.dt.tzinfo)
        return True

    def advance_month(self) -> None:
        """Roll forward the month component until it satisfies the constraints."""

        if self.dt.month in self.months:
            return

        needle = self.dt.date()
        while needle.month not in self.months:
            needle = (needle.replace(day=1) + td(days=32)).replace(day=1)

        self.dt = datetime.combine(needle, time(), tzinfo=self.dt.tzinfo)

    def __iter__(self) -> "CronSim":
        return self

    def __next__(self) -> datetime:
        self.tick()

        start_year = self.dt.year
        while True:
            self.advance_month()
            if self.dt.year > start_year + 50:
                # Give up if there is no match for 50 years.
                # It would be nice to detect "this will never yield any results"
                # situations in a more intelligent way.
                raise StopIteration

            if self.advance_day():
                continue

            if self.advance_hour():
                continue

            if self.advance_minute():
                continue

            # If all constraints are satisfied then we have the result.
            # The last step is to check if we need to fix up an imaginary
            # or ambiguous date.
            if self.fixup_tz:
                result = self.dt.replace(tzinfo=self.fixup_tz, fold=0)
                while is_imaginary(result):
                    self.dt += td(minutes=1)
                    result = self.dt.replace(tzinfo=self.fixup_tz)

                return result

            return self.dt

    def explain(self) -> str:
        from cronsim.explain import Expression

        return Expression(self.parts).explain()



from __future__ import annotations

from dataclasses import dataclass
from typing import Generator


@dataclass(frozen=True)
class Sequence:
    start: int | None = None
    stop: int | None = None
    step: int = 1
    nth: int | None = None

    def is_star(self) -> bool:
        """Return True if this sequence describes a wildcard."""
        return self.start is None and self.step == 1

    def is_single(self) -> bool:
        """Return True if this sequence describes a a single, specific value."""
        return self.start is not None and self.start == self.stop


def join(l: list[str]) -> str:
    """Join together strings using commas and 'and' as separators.

    >>> join(["a"])
    'a'
    >>> join(["a", "b"])
    'a and b'
    >>> join(["a", "b", "c"])
    'a, b, and c'
    """
    if len(l) == 1:
        return l[0]

    if len(l) == 2:
        return f"{l[0]} and {l[1]}"

    head = ", ".join(l[:-1])
    return f"{head}, and {l[-1]}"


ORDINALS = {
    -1: "last",
    1: "first",
    2: "second",
    3: "third",
    4: "fourth",
    5: "fifth",
    6: "sixth",
    7: "seventh",
    8: "eighth",
    9: "ninth",
}


def ordinal(x: int) -> str:
    """Format integer as an ordinal number.

    >>> ordinal(1)
    'first'
    >>> ordinal(15)
    '15th'
    """
    return ORDINALS.get(x, f"{x}th")


def format_time(h: int, m: int) -> str:
    """Format hours and minutes as HH:MM.

    >>> format_time(0, 0)
    '00:00'
    >>> format_time(1, 23)
    '01:23'
    """
    return f"{h:02d}:{m:02d}"


class Field(object):
    name = "FIXME"
    symbolic: list[str] = []
    min_value = 0
    max_value = 0

    def __init__(self, value: str):
        self.value = value
        self.parsed: list[Sequence] = []

        for seq in self.parse(value):
            if seq not in self.parsed:
                self.parsed.append(seq)

        # If the field contains a single numeric value,
        # store it in self.single_value
        self.single_value = self.parsed[0].start
        for seq in self.parsed:
            if seq.start != self.single_value or seq.stop != self.single_value:
                self.single_value = None
                break

        # Does the field cover the full range?
        self.star = all(seq.is_star() for seq in self.parsed)
        # Are there any single values?
        self.any_singles = any(seq.is_single() for seq in self.parsed)
        # Are all values single values?
        self.all_singles = all(seq.is_single() for seq in self.parsed)

    def parse(self, value: str) -> Generator[Sequence, None, None]:
        """Parse a single field of a cron expression into Sequence objects."""
        for term in value.split(","):
            if term == "*":
                yield Sequence()
            elif isinstance(self, Weekday) and term.endswith("L"):
                v = self._int(term[:-1])
                yield Sequence(start=v, stop=v, nth=-1)
            elif "#" in term:
                term, nth = term.split("#")
                v = self._int(term)
                yield Sequence(start=v, stop=v, nth=int(nth))
            elif "/" in term:
                term, step_str = term.split("/")
                step = int(step_str)
                if term == "*":
                    yield Sequence(step=step)
                elif term == "LW":
                    yield Sequence(start=-2, stop=-2, nth=-1)
                elif term == "L":
                    yield Sequence(start=-1, stop=-1, nth=-1)
                else:
                    if "-" in term:
                        start_str, stop_str = term.split("-")
                        start, stop = self._int(start_str), self._int(stop_str)
                    else:
                        start = self._int(term)
                        stop = self.max_value

                    if start <= self.min_value and stop >= self.max_value:
                        yield Sequence(step=step)
                    else:
                        yield Sequence(start=start, stop=stop, step=step)
            elif "-" in term:
                start_str, stop_str = term.split("-")
                start, stop = self._int(start_str), self._int(stop_str)
                if start + 1 == stop:
                    # treat a 2-long sequence as two single values:
                    yield Sequence(start=start, stop=start)
                    yield Sequence(start=stop, stop=stop)
                elif start <= self.min_value and stop >= self.max_value:
                    yield Sequence()
                else:
                    yield Sequence(start=start, stop=stop)
            elif term == "LW":
                yield Sequence(start=-2, stop=-2, nth=-1)
            elif term == "L":
                yield Sequence(start=-1, stop=-1, nth=-1)
            else:
                v = self._int(term)
                yield Sequence(start=v, stop=v)

    def _int(self, value: str) -> int:
        """Convert a value from a cron expression to an integer.

        For month and weekday fields, this takes care of converting JAN, FEB, ...
        and MON, TUE, ...
        """
        if value in self.symbolic:
            return self.symbolic.index(value)
        return int(value)

    def singles(self) -> list[int]:
        """Return a list only the single values in this field."""
        return [seq.start for seq in self.parsed if isinstance(seq.start, int)]

    def label(self, idx: int) -> str:
        """Convert an integer value to a string for display.

        >>> label(1)
        '1'
        """
        return str(idx)

    def format_single(self, value: int) -> str:
        """Format a single value for display.

        >>> format_single(1)
        'minute 1'
        """
        return f"{self.name} {self.label(value)}"

    def format_nth(self, value: int, nth: int) -> str:
        """Format nth-weekday-of-month and L values.

        Implemented in Month and Weekday subclasses.
        """
        raise NotImplementedError

    def format_every(self, step: int = 1) -> str:
        """Format wildcard and wildcard-with-step values.

        >>> format_every(1)
        'every minute'

        >>> format_every(5)
        'every 5th minute'
        """
        if step == 1:
            return f"every {self.name}"
        return f"every {ordinal(step)} {self.name}"

    def format_seq(self, start: int, stop: int, step: int = 1) -> str:
        """Format a sequence.

        >>> format_seq(1, 10)
        'every minute from 1 through 10'

        >>> format_seq(1, 10, 2)
        'every 2nd minute from 1 through 10'
        """
        start_str = self.label(start)
        stop_str = self.label(stop)
        if step == 1:
            return f"every {self.name} from {start_str} through {stop_str}"
        return f"every {ordinal(step)} {self.name} from {start_str} through {stop_str}"

    def format(self) -> str:
        """Format every component of this field, and join them together."""
        parts = []
        for seq in self.parsed:
            if seq.start is None:
                parts.append(self.format_every(seq.step))
            elif seq.stop != seq.start:
                assert seq.stop is not None
                parts.append(self.format_seq(seq.start, seq.stop, seq.step))
            elif seq.nth is not None:
                parts.append(self.format_nth(seq.start, seq.nth))
            else:
                parts.append(self.format_single(seq.start))

        return join(parts)

    def __str__(self) -> str:
        """Return a human-friendly representation of this field.

        Subclasses can override this method to add prepositions ("at", "on", "in").
        """
        return self.format()


class Minute(Field):
    name = "minute"
    min_value = 0
    max_value = 59

    def format(self) -> str:
        """Format the minute field.

        This method adds special handling when all values are single values. Instead
        of "minute 1, minute 3, and minute 5", it produces "minutes 1, 3, and 5".
        """
        if self.all_singles and len(self.parsed) > 1:
            labels = [self.label(v) for v in self.singles()]
            return f"minutes {join(labels)}"
        return super().format()

    def __str__(self) -> str:
        """Return a human-friendly representation of the minute field.

        This method adds the 'at' preposition to the formatted string if the field has
        any single values.

        For example, if the field has a single range (1-5), the result will be
        "every minute from 1 through 5". But if the field also has a single field
        (1-5,10), the result will be "at every minute from 1 through 5 and minute 10".
        """
        result = super().__str__()
        if self.any_singles:
            return "at " + result

        return result


class Hour(Field):
    name = "hour"
    min_value = 0
    max_value = 23

    def format(self) -> str:
        """Format the hour field.

        This method adds special handling when all values are single values. Instead
        of "hour 1, hour 3, and hour 5", it produces "hours 1, 3, and 5".
        """
        if self.all_singles and len(self.parsed) > 1:
            labels = [self.label(v) for v in self.singles()]
            return f"hours {join(labels)}"
        return super().format()

    def __str__(self) -> str:
        """Return a human-friendly representation of the minute field.

        This method adds the 'past' preposition to the formatted string. For example,
        if the hour field has a single value (5), the result will be "past hour 5".
        """
        return "past " + super().__str__()


class Day(Field):
    name = "day of month"
    min_value = 1
    max_value = 31

    def format_single(self, value: int) -> str:
        """Format a single day-of-month value for display.

        >>> format_single(2)
        'the 2nd day of month'
        """
        return f"the {ordinal(value)} day of month"

    def format_nth(self, value: int, nth: int) -> str:
        """Format the L and LW values.

        >>> format_nth(-1)
        'the last day of the month'
        >>> format_nth(-2)
        'the last weekday of the month'
        """
        if nth == -1 and value == -1:
            return "the last day of the month"
        if nth == -1 and value == -2:
            return "the last weekday of the month"
        return super().format_nth(value, nth)

    def format(self) -> str:
        """Format the day field.

        This method adds special handling for the case when
        all values are single values. For example, instead of
        "the 1st day of month, the 3rd day of month, and the 5th day of month"
        it produces "the 1st, 3rd, and 5th day of month".

        We cannot apply this optimization if the single values contain "-2"
        (the special value for "last weekday of the month").
        """
        if self.all_singles and len(self.parsed) > 1:
            singles = self.singles()
            if -2 not in singles:
                labels = [f"the {ordinal(v)}" for v in singles]
                return f"{join(labels)} day of month"
        return super().format()

    def __str__(self) -> str:
        """Return a human-friendly representation of the day of month field.

        This method unconditionally adds the 'on' preposition.
        """
        return "on " + super().__str__()


class Month(Field):
    name = "month"
    min_value = 1
    max_value = 12
    symbolic = "_ JAN FEB MAR APR MAY JUN JUL AUG SEP OCT NOV DEC".split()
    labels = "_ January February March April May June July August September October November December".split()

    def label(self, idx: int) -> str:
        """Convert an integer value to a month name.

        >>> label(1)
        'January'
        """
        return self.labels[idx]

    def format_single(self, value: int) -> str:
        """Format a single month value for display.

        >>> format_single(2)
        'February'
        """
        return self.label(value)

    def __str__(self) -> str:
        """Return a human-friendly representation of the month field.

        This method unconditionally adds the 'in' preposition.
        """
        return "in " + super().__str__()


class Weekday(Field):
    name = "day of week"
    min_value = 0
    max_value = 7
    symbolic = "SUN MON TUE WED THU FRI SAT SUN".split()
    labels = "Sunday Monday Tuesday Wednesday Thursday Friday Saturday Sunday".split()

    def label(self, idx: int) -> str:
        """Convert an integer value to a day of week name.

        >>> label(1)
        'Monday'
        """
        return self.labels[idx]

    def format_single(self, value: int) -> str:
        """Format a single month value for display.

        >>> format_single(2)
        'Tuesday'
        """
        return self.label(value)

    def format_nth(self, value: int, nth: int) -> str:
        """Format the nth-weekday-of-month value.

        >>> format_nth(2, -1)
        'the last Tuesday of the month'
        >>> format_nth(2, 4)
        'the 4th Tuesday of the month'
        """
        label = self.label(value)
        if nth == -1:
            return f"the last {label} of the month"

        return f"the {ordinal(nth)} {label} of the month"

    def format_seq(self, start: int, stop: int, step: int = 1) -> str:
        """Format a sequence of weekdays.

        >>> format_seq(1, 3)
        'Monday through Wednesday'
        >>> format_seq(1, 7, 2)
        'evary 2nd day of week from Monday through Sunday'
        """

        if step == 1:
            # "Monday through Friday"
            # instead of "every day of week from Monday through Friday"
            return f"{self.label(start)} through {self.label(stop)}"

        return super().format_seq(start, stop, step)

    def __str__(self) -> str:
        """Return a human-friendly representation of the month field.

        This method unconditionally adds the 'on' preposition.
        """
        return "on " + super().__str__()


class Expression(object):
    def __init__(self, parts: list[str]):
        self.minute = Minute(parts[0])
        self.hour = Hour(parts[1])
        self.day = Day(parts[2])
        self.month = Month(parts[3])
        self.dow = Weekday(parts[4])
        self.day_and = parts[2].startswith("*") or parts[4].startswith("*")

    def optimized_times(self) -> tuple[str, bool] | None:
        """Apply formatting optimizations for hours and minutes.

        If both hours and minutes contain only a few single values, format them
        as "at HH:MM, HH:MM, HH:MM, and HH:MM"

        If hours have a single value, and minutes have a single sequence with step 1,
        format them as "every minute from HH:MM through HH:MM".

        If minutes have a single "*/<step>" sequence and hours have a sequence
        with step 1, format them as "every nth minute from HH:00 through HH:59"

        If no special optimizations apply, return None.
        """
        # at 11:00, 11:30, ...
        if self.hour.all_singles and self.minute.all_singles:
            minute_terms, hour_terms = self.minute.singles(), self.hour.singles()

            if len(minute_terms) * len(hour_terms) <= 4:
                times = []
                for h in sorted(hour_terms):
                    for m in sorted(minute_terms):
                        times.append(format_time(h, m))

                return "at " + join(times), True

        # every minute from 11:00 through 11:10
        if self.hour.single_value and len(self.minute.parsed) == 1:
            seq = self.minute.parsed[0]
            if seq.start is not None and seq.stop is not None and seq.step == 1:
                hhmm1 = format_time(self.hour.single_value, seq.start)
                hhmm2 = format_time(self.hour.single_value, seq.stop)
                return f"every minute from {hhmm1} through {hhmm2}", False

        # every minute from 9:00 through 17:59
        if len(self.hour.parsed) == 1 and len(self.minute.parsed) == 1:
            mseq = self.minute.parsed[0]
            if mseq.start is None:
                hseq = self.hour.parsed[0]
                if hseq.start is not None and hseq.stop is not None and hseq.step == 1:
                    hhmm1 = format_time(hseq.start, 0)
                    hhmm2 = format_time(hseq.stop, 59)
                    return f"{self.minute} from {hhmm1} through {hhmm2}", False

        return None

    def optimized_dates(self) -> str | None:
        """Apply formatting optimizations for specific dates.

        If day-of-month is L, format it as
        "on the last day of <month description here>".

        If day-of-month is LW, format it as
        "on the last weekday of <month description here>".

        If month and day-of-month each have a single value
        (for example, month 2 and day-of-month 1), format them as
        "February 1st".

        If day-of-month has a single value (for example, 2),
        format it as "on the 2nd day of <month description here>".

        If no special optimizations apply, return None.
        """

        if self.dow.star:
            if self.day.single_value == -1:
                return f"on the last day of {self.month.format()}"
            if self.day.single_value == -2:
                return f"on the last weekday of {self.month.format()}"
            if self.month.single_value and self.day.single_value:
                return f"on {self.month.format()} {self.day.single_value}"
            if self.day.single_value:
                date_ord = ordinal(self.day.single_value)
                return f"on the {date_ord} day of {self.month.format()}"

        return None

    def translate_time(self) -> tuple[str, bool]:
        """Convert the minute and hour fields to text."""
        if self.hour.star:
            if self.minute.star:
                return "every minute", False
            if self.minute.single_value == 0:
                return "at the start of every hour", False
            return f"{self.minute} of every hour", False

        if times := self.optimized_times():
            return times

        return f"{self.minute} {self.hour}", False

    def translate_date(self, allow_every_day: bool) -> str:
        """Convert the day, month, and weekday fields to text."""
        if dates := self.optimized_dates():
            return dates

        if allow_every_day:
            if self.day.star and self.dow.star and self.month.all_singles:
                # At ... every day in January
                return f"every day {self.month}"

        parts: list[Field | str] = []
        if not self.day.star:
            parts.append(self.day)
        if not self.dow.star:
            if not self.day.star and self.day_and:
                parts.append("if it's")
            elif not self.day.star and not self.day_and:
                parts.append("and")
            parts.append(self.dow)
        if not self.month.star:
            parts.append(self.month)

        return " ".join(str(part) for part in parts)

    def explain(self) -> str:
        """Convert the full expression to text."""
        time, allow_every_day = self.translate_time()
        if date := self.translate_date(allow_every_day):
            result = f"{time} {date}"
        elif allow_every_day:
            result = f"{time} every day"
        else:
            result = f"{time}"

        return result[0].upper() + result[1:]


def explain(expr: str) -> str:
    """Convert the given cron expression to human-friendly description.

    >>> explain("0 0 15 JAN-FEB *")
    'At 00:00 on the 15th day of January and February'
    """
    parts = expr.upper().split()
    return Expression(parts).explain()


# For quick testing, you can run this script from shell like so:
# python explain.py "0 0 15 JAN-FEB *"
if __name__ == "__main__":
    import sys

    if len(sys.argv) == 2:
        print(explain(sys.argv[1]))
